@@ -1,6 +1,6 @@
 /* 
    Socket handling tests
-   Copyright (C) 2002-2006, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 2002-2011, Joe Orton <joe@manyfish.co.uk>
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -32,6 +32,9 @@
 #ifdef HAVE_STRING_H
 #include <string.h>
 #endif
+#ifdef HAVE_UNISTD_H
+#include <unistd.h> /* for gethostname() */
+#endif
 #include <time.h> /* for time() */
 
 #include "ne_socket.h"
@@ -168,10 +171,11 @@ static int wrap_serve(ne_socket *sock, void *ud)
 static int begin(ne_socket **sock, server_fn fn, void *ud)
 {
     struct serve_pair pair;
+    unsigned int port;
     pair.fn = fn;
     pair.userdata = ud;
-    CALL(spawn_server(7777, wrap_serve, &pair));
-    CALL(do_connect(sock, localhost, 7777));
+    CALL(new_spawn_server(1, wrap_serve, &pair, &port));
+    CALL(do_connect(sock, localhost, port));
     ONV(ne_sock_connect_ssl(*sock, client_ctx, NULL),
 	("SSL negotation failed: %s", ne_sock_error(*sock)));
     return OK;
@@ -181,8 +185,9 @@ static int begin(ne_socket **sock, server_fn fn, void *ud)
 /* non-SSL begin() function. */
 static int begin(ne_socket **sock, server_fn fn, void *ud)
 {
-    CALL(spawn_server(7777, fn, ud));
-    return do_connect(sock, localhost, 7777);
+    unsigned int port;
+    CALL(new_spawn_server(1, fn, ud, &port));
+    return do_connect(sock, localhost, port);
 }
 #endif
 
@@ -214,13 +219,28 @@ static int resolve_ipv6(void)
 #endif
 
 static const unsigned char raw_127[4] = "\x7f\0\0\01", /* 127.0.0.1 */
-raw6_nuls[16] = /* :: */ "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
+    raw6_nuls[16] = /* :: */ "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0";
 #ifdef TEST_IPV6
 static const unsigned char 
+raw6_local[16] = "\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\1",
 raw6_cafe[16] = /* feed::cafe */ "\xfe\xed\0\0\0\0\0\0\0\0\0\0\0\0\xca\xfe",
 raw6_babe[16] = /* cafe:babe:: */ "\xca\xfe\xba\xbe\0\0\0\0\0\0\0\0\0\0\0\0";
 #endif
 
+/* Check the given inet addr is 127.0.0.1. */
+static int check_is_raw127(const ne_inet_addr *ia)
+{
+    unsigned char raw[5];
+
+    raw[4] = 'Z';
+    ONN("bogus ne_iaddr_typeof return", ne_iaddr_typeof(ia) != ne_iaddr_ipv4);
+    ONN("ne_iaddr_raw gave bad retval", ne_iaddr_raw(ia, raw) != raw);
+    ONN("raw address mismatch", memcmp(raw, raw_127, 4) != 0);
+    ONN("ne_iaddr_raw buffer overflow", raw[4] != 'Z');
+
+    return OK;
+}
+
 static int addr_make_v4(void)
 {
     ne_inet_addr *ia;
@@ -232,9 +252,24 @@ static int addr_make_v4(void)
     ne_iaddr_print(ia, pr, sizeof pr);
     ONV(strcmp(pr, "127.0.0.1"), ("address was %s not 127.0.0.1", pr));
 
-    ONN("bogus ne_iaddr_typeof return", ne_iaddr_typeof(ia) != ne_iaddr_ipv4);
+    CALL(check_is_raw127(ia));
+    
+    ne_iaddr_free(ia);
+
+    return OK;
+}
+
+static int parse_v4(void)
+{
+    ne_inet_addr *ia;
+
+    ia = ne_iaddr_parse("127.0.0.1", ne_iaddr_ipv4);
+    ONN("parse failed", ia == NULL);
+
+    CALL(check_is_raw127(ia));
 
     ne_iaddr_free(ia);
+
     return OK;
 }
 
@@ -255,22 +290,40 @@ static int addr_make_v6(void)
     for (n = 0; as[n].rep != NULL; n++) {
 	ne_inet_addr *ia = ne_iaddr_make(ne_iaddr_ipv6, as[n].addr);
 	char pr[128];
+        unsigned char raw[17];
+
+	ONV(ia == NULL, ("could not make address for '%s'", 
+                         as[n].rep));
 
-	ONV(ia == NULL, ("could not make address for %d", n));
 	ne_iaddr_print(ia, pr, sizeof pr);
 	ONV(strcmp(pr, as[n].rep), 
 	    ("address %d was '%s' not '%s'", n, pr, as[n].rep));
 	
         ONN("bogus ne_iaddr_typeof return", ne_iaddr_typeof(ia) != ne_iaddr_ipv6);
 
+        raw[16] = 'Z';
+        ONN("ne_iaddr_raw gave bad retval", ne_iaddr_raw(ia, raw) != raw);
+        ONN("raw address mismatch", memcmp(raw, as[n].addr, 4) != 0);
+        ONN("ne_iaddr_raw buffer overflow", raw[16] != 'Z');
+        
 	ne_iaddr_free(ia);
-    }
+        
+        ia = ne_iaddr_parse(as[n].rep, ne_iaddr_ipv6);
+        ONV(ia == NULL, ("ne_iaddr_parse failed for %s", as[n].rep));
+        ONN("bogus ne_iaddr_typeof return", ne_iaddr_typeof(ia) != ne_iaddr_ipv6);
+        ONN("ne_iaddr_raw gave bad retval", ne_iaddr_raw(ia, raw) != raw);
+        ONN("raw address mismatch", memcmp(raw, as[n].addr, 4) != 0);
+        ONN("ne_iaddr_raw buffer overflow", raw[16] != 'Z');
+
+        ne_iaddr_free(ia);
+   }
 
     return OK;
 #else
     /* should fail when lacking IPv6 support. */
     ne_inet_addr *ia = ne_iaddr_make(ne_iaddr_ipv6, raw6_nuls);
     ONN("ne_iaddr_make did not return NULL", ia != NULL);
+    ONN("ne_iaddr_parse did not return NULL", ne_iaddr_parse("127.0.0.1", ne_iaddr_ipv6));
 #endif
     return OK;
 }
@@ -297,6 +350,7 @@ static int addr_compare(void)
 #ifdef TEST_IPV6
     ne_iaddr_free(ia2);
     ia2 = ne_iaddr_make(ne_iaddr_ipv6, raw6_cafe);
+    ONN("could not make IPv6 address", !ia2);
     ret = ne_iaddr_cmp(ia1, ia2);
     ONN("comparison of IPv4 and IPv6 addresses was zero", ret == 0);
 
@@ -316,6 +370,52 @@ static int addr_compare(void)
     return OK;
 }
 
+static int addr_reverse(void)
+{
+    ne_inet_addr *ia = ne_iaddr_make(ne_iaddr_ipv4, raw_127);
+    char buf[128], *syshost = NULL;
+
+#ifdef HAVE_GETHOSTNAME
+    char host[128];
+
+    if (gethostname(host, sizeof host) == 0) {
+        syshost = host;
+    }
+#endif
+
+    ONN("ne_iaddr_make returned NULL", ia == NULL);
+
+    ONN("reverse lookup for 127.0.0.1 failed",
+        ne_iaddr_reverse(ia, buf, sizeof buf) != 0);
+
+    ONV(!(strcmp(buf, "localhost.localdomain") == 0
+          || strcmp(buf, "localhost") == 0
+          || (syshost && strcmp(buf, syshost) == 0)),
+        ("reverse lookup for 127.0.0.1 got %s", buf));
+
+    ne_iaddr_free(ia);
+
+    return OK;
+}
+
+static int addr_canonical(void)
+{
+    ne_sock_addr *sa;
+    const char *h;
+
+    sa = ne_addr_resolve("localhost", NE_ADDR_CANON);
+    ONN("could not resolve localhost", sa == NULL);
+    
+    h = ne_addr_canonical(sa);
+    ONN("no canonical name for localhost", h == NULL);
+
+    NE_DEBUG(NE_DBG_SOCKET, "canonical name: %s\n", h);
+
+    ne_addr_destroy(sa);
+    
+    return OK;
+}
+
 static int just_connect(void)
 {
     ne_socket *sock;
@@ -331,12 +431,13 @@ static int addr_connect(void)
 {
     ne_socket *sock = ne_sock_create();
     ne_inet_addr *ia;
+    unsigned int port;
 
     ia = ne_iaddr_make(ne_iaddr_ipv4, raw_127);
     ONN("ne_iaddr_make returned NULL", ia == NULL);
     
-    CALL(spawn_server(7777, serve_close, NULL));
-    ONN("could not connect", ne_sock_connect(sock, ia, 7777));
+    CALL(new_spawn_server(1, serve_close, NULL, &port));
+    ONN("could not connect", ne_sock_connect(sock, ia, port));
     ne_sock_close(sock);
     CALL(await_server());
 
@@ -344,6 +445,34 @@ static int addr_connect(void)
     return OK;
 }
 
+static int addr_peer(void)
+{
+    ne_socket *sock = ne_sock_create();
+    ne_inet_addr *ia, *ia2;
+    unsigned int port = 9999, realport;
+    int ret;
+
+    ia = ne_iaddr_make(ne_iaddr_ipv4, raw_127);
+    ONN("ne_iaddr_make returned NULL", ia == NULL);
+    
+    CALL(new_spawn_server(1, serve_close, NULL, &realport));
+    ONN("could not connect", ne_sock_connect(sock, ia, realport));
+
+    ia2 = ne_sock_peer(sock, &port);
+    ret = ne_iaddr_cmp(ia, ia2);
+    ONV(ret != 0,
+        ("comparison of peer with server address was %d", ret));
+
+    ONV(port != realport, ("got peer port %u, expected %u", port, realport));
+ 
+    ne_sock_close(sock);
+    CALL(await_server());
+
+    ne_iaddr_free(ia);
+    ne_iaddr_free(ia2);
+    return OK;
+}
+
 /* Exect a read() to return EOF */
 static int expect_close(ne_socket *sock)
 {
@@ -446,6 +575,7 @@ static int fullread_expect(ne_socket *sock, const char *str, size_t len)
     return OK;
 }
 
+#define FULLREAD(str) CALL(fullread_expect(sock, str, strlen(str)))
 
 /* Declare a struct string */
 #define DECL(var,str) struct string var = { str, 0 }; var.len = strlen(str)
@@ -496,7 +626,6 @@ static int small_reads(void)
 
 /* peek or read, expecting to get given string. */
 #define READ(str) CALL(read_expect(sock, str, strlen(str)))
-#define FULLREAD(str) CALL(fullread_expect(sock, str, strlen(str)))
 #define PEEK(str) CALL(peek_expect(sock, str, strlen(str)))
 
 /* Stress out the read buffer handling a little. */
@@ -542,12 +671,11 @@ static int line_expect(ne_socket *sock, const char *line)
 {
     ssize_t ret = ne_sock_readline(sock, buffer, BUFSIZ);
     size_t len = strlen(line);
+    NE_DEBUG(NE_DBG_SOCKET, " -> expected=%s -> actual=%s", line, buffer);
     ONV(ret == NE_SOCK_CLOSED, ("socket closed, expecting `%s'", line));
     ONV(ret < 0, ("socket error `%s', expecting `%s'", 
 		  ne_sock_error(sock), line));
-    ONV((size_t)ret != len, 
-	("readline got %" NE_FMT_SSIZE_T ", expecting `%s'", ret, line));
-    ONV(strcmp(line, buffer),
+    ONV((size_t)ret != len || strcmp(line, buffer),
 	("readline mismatch: `%s' not `%s'", buffer, line));
     return OK;
 }
@@ -704,10 +832,35 @@ static int to_end(ne_socket *sock)
 	 
 #define TO_BEGIN ne_socket *sock; CALL(to_begin(&sock))
 #define TO_OP(x) do { int to_ret = (x); \
-ONV(to_ret != NE_SOCK_TIMEOUT, ("operation did not timeout: %d", to_ret)); \
+        ONV(to_ret != NE_SOCK_TIMEOUT, ("operation did not timeout: got %d (%s)", to_ret, ne_sock_error(sock))); \
 } while (0)
 #define TO_FINISH return to_end(sock)
 
+#ifndef TEST_CONNECT_TIMEOUT
+#define TEST_CONNECT_TIMEOUT 0
+#endif
+#if TEST_CONNECT_TIMEOUT
+
+/* No obvious way to reliably test a connect() timeout.  But
+ * www.example.com seems to drop packets on ports other than 80 so
+ * that actually works pretty well.  Disabled by default. */
+static int connect_timeout(void)
+{
+    static const unsigned char example_dot_com[] = "\xC0\x00\x22\xA6";
+    ne_socket *sock = ne_sock_create();
+    ne_inet_addr *ia = ne_iaddr_make(ne_iaddr_ipv4, example_dot_com);
+
+    ne_sock_connect_timeout(sock, 1);
+
+    TO_OP(ne_sock_connect(sock, ia, 8080));
+
+    ne_iaddr_free(ia);
+    ne_sock_close(sock);
+
+    return OK;
+}
+#endif
+
 static int peek_timeout(void)
 {
     TO_BEGIN;
@@ -756,13 +909,6 @@ static int serve_expect(ne_socket *sock, void *ud)
     return OK;
 }
 
-static int full_write(ne_socket *sock, const char *data, size_t len)
-{
-    int ret = ne_sock_fullwrite(sock, data, len);
-    ONV(ret, ("write failed (%d): %s", ret, ne_sock_error(sock)));
-    return OK;
-}
-
 #define WRITEL(str) CALL(full_write(sock, str, strlen(str))); \
 minisleep()
 
@@ -798,6 +944,43 @@ static int large_writes(void)
     return finish(sock, 1);    
 }
 
+static int full_writev(ne_socket *sock, struct ne_iovec *vec, int count)
+{
+    int ret = ne_sock_fullwritev(sock, vec, count);
+    NE_DEBUG(NE_DBG_SOCKET, "wrote vector (%d)\n", count);
+    ONV(ret, ("writev failed (%d): %s", ret, ne_sock_error(sock)));
+    return OK;
+}
+
+#undef LARGE_SIZE
+#define LARGE_SIZE (123456 * 4)
+
+static int large_writev(void)
+{
+    struct string str;
+    ne_socket *sock;
+    ssize_t n;
+    struct ne_iovec vec[4];
+
+    str.data = ne_malloc(LARGE_SIZE);
+    str.len = LARGE_SIZE;
+
+    for (n = 0; n < LARGE_SIZE; n++)
+	str.data[n] = 41 + n % 130;
+    
+    for (n = 0; n < 4; n++) {
+        vec[n].base = str.data + n * LARGE_SIZE / 4;
+        vec[n].len = LARGE_SIZE / 4;
+    }
+
+    CALL(begin(&sock, serve_expect, &str));
+    CALL(full_writev(sock, vec, 4));
+    
+    ne_free(str.data);
+    return finish(sock, 1);    
+}
+
+
 /* echoes back lines. */
 static int echo_server(ne_socket *sock, void *ud)
 {
@@ -847,7 +1030,8 @@ static int ssl_closure(void)
     ONV(ret != NE_SOCK_RESET && ret != NE_SOCK_CLOSED, 
 	("write got %" NE_FMT_SSIZE_T " not reset or closure: %s", ret,
          ne_sock_error(sock)));
-    return good_close(sock);
+    ne_sock_close(sock);
+    return OK;
 }
 
 static int serve_truncate(ne_socket *sock, void *userdata)
@@ -866,7 +1050,9 @@ static int ssl_truncate(void)
     ONV(ret != NE_SOCK_TRUNC,
 	("socket got error %d not truncation: `%s'", ret,
 	 ne_sock_error(sock)));
-    return finish(sock, 0);
+    ne_sock_close(sock);
+    CALL(await_server());
+    return OK;
 }
 
 #else
@@ -874,6 +1060,7 @@ static int ssl_truncate(void)
 /* use W Richard Stevens' SO_LINGER trick to elicit a TCP RST */
 static int serve_reset(ne_socket *sock, void *ud)
 {
+    minisleep();
     reset_socket(sock);
     exit(0);
     return 0;
@@ -891,8 +1078,12 @@ static int write_reset(void)
         ne_sock_close(sock);
         return SKIP;
     }
-    ONV(ret != NE_SOCK_RESET, 
-        ("write got %d not reset: %s", ret, ne_sock_error(sock)));
+    if (ret == NE_SOCK_CLOSED) {
+        t_warning("got EOF, failed to elicit TCP RST");
+    } else {
+        ONV(ret != NE_SOCK_RESET, 
+            ("write got %d not reset: %s", ret, ne_sock_error(sock)));
+    }
     return good_close(sock);
 }
 
@@ -971,6 +1162,14 @@ static int ssl_session_id(void)
     ONN("retrieve session id length",
         ne_sock_sessid(sock, NULL, &len1));
 
+    if (len1 == 0) {
+        /* recent versions of OpenSSL seem to do this, not sure
+         * why or whether it's bad. */
+        finish(sock, 1);
+        t_context("zero-length session ID, cannot test further");
+        return SKIP;
+    }            
+
     if (len1 < sizeof buf) {
         buf[len1] = 'Z';
     }
@@ -978,14 +1177,14 @@ static int ssl_session_id(void)
     {
         size_t len2;
 
-        len2 = len1;
+        len2 = sizeof buf;
         ONN("could not retrieve session id",
             ne_sock_sessid(sock, buf, &len2));
         
         ONN("buffer size changed!?", len1 != len2);
     }        
 
-    ONN("buffer written past end", 
+    ONN("buffer written past expected end", 
         len1 < sizeof buf && buf[len1] != 'Z');
 
     /* Attempt retrieval into too-short buffer: */
@@ -1003,6 +1202,262 @@ static int ssl_session_id(void)
     return await_server();
 }
 
+static int serve_ppeer(ne_socket *sock, void *ud)
+{
+    unsigned int port = 99999;
+    ne_inet_addr *ia = ne_sock_peer(sock, &port);
+    char buf[128], line[256];
+
+    if (ia == NULL)
+        ne_snprintf(line, sizeof line, "error: %s", ne_sock_error(sock));
+    else
+        ne_snprintf(line, sizeof line,
+                    "%s@%u\n", ne_iaddr_print(ia, buf, sizeof buf),
+                    port);
+
+    CALL(full_write(sock, line, strlen(line)));
+         
+    ne_iaddr_free(ia);
+    
+    return OK;
+}
+
+static int try_prebind(int addr, int port)
+{
+    ne_socket *sock = ne_sock_create();
+    ne_inet_addr *ia;
+    char buf[128], line[256];
+    unsigned int srvport;
+
+    ia = ne_iaddr_make(ne_iaddr_ipv4, raw_127);
+    ONN("ne_iaddr_make returned NULL", ia == NULL);
+    
+    CALL(new_spawn_server(1, serve_ppeer, NULL, &srvport));
+
+    ne_sock_prebind(sock, addr ? ia : NULL, port ? 7778 : 0);
+
+    ONN("could not connect", ne_sock_connect(sock, ia, srvport));
+
+    ne_snprintf(line, sizeof line,
+                "%s@%d\n", ne_iaddr_print(ia, buf, sizeof buf),
+                7778);
+    
+    if (!port) {
+        /* Don't know what port will be chosen, so... */
+        ssize_t ret = ne_sock_readline(sock, buffer, BUFSIZ);
+        
+        ONV(ret < 0, ("socket error `%s'", ne_sock_error(sock)));
+
+        ONV(strncmp(line, buffer, strchr(line, '@') - line) != 0,
+            ("bad address: '%s', expecting '%s'",
+             buffer, line));
+    }
+    else {
+        LINE(line);
+    }
+
+    ne_sock_close(sock);
+    CALL(await_server());
+
+    ne_iaddr_free(ia);
+    return OK;
+}
+
+static int prebind(void)
+{
+    CALL(try_prebind(1, 0));
+    CALL(try_prebind(0, 1));
+    CALL(try_prebind(1, 1));
+
+    return OK;
+}
+
+static int serve_cipher(ne_socket *sock, void *ud)
+{
+    char *ciph = ne_sock_cipher(sock);
+    char *s = ciph && strlen(ciph) ? ciph : "NULL";
+
+    CALL(full_write(sock, s, strlen(s)));
+
+    if (ciph) ne_free(ciph);
+    
+    return OK;
+}
+
+static int cipher(void)
+{
+    ne_socket *sock;
+
+#ifdef SOCKET_SSL
+    char *ciph;
+
+    CALL(begin(&sock, serve_cipher, NULL));
+
+    ciph = ne_sock_cipher(sock);
+
+    ONN("NULL/empty cipher", ciph == NULL || strlen(ciph) == 0);
+
+    FULLREAD(ciph);
+    
+    ne_free(ciph);
+    
+#else
+    CALL(begin(&sock, serve_cipher, NULL));
+
+    ONN("non-NULL cipher for non-SSL socket", 
+        ne_sock_cipher(sock) != NULL);
+
+    FULLREAD("NULL");
+
+#endif
+    return finish(sock, 1);
+}
+
+static int error(void)
+{
+    ne_socket *sock = ne_sock_create();
+
+    ne_sock_set_error(sock, "%s:%s", "fish", "42");
+    
+    ONCMP("fish:42", ne_sock_error(sock), "socket error", "set");
+
+    ne_sock_close(sock);
+    return OK;
+}
+
+static int begin_socks(ne_socket **sock, struct socks_server *srv,
+                       server_fn server, void *userdata)
+{
+    unsigned int port;
+    srv->server = server;
+    srv->userdata = userdata;
+    srv->say_hello = 1;
+    CALL(new_spawn_server(1, socks_server, srv, &port));
+    return do_connect(sock, localhost, port);
+}
+
+static int socks_proxy(void)
+{
+    static const struct {
+        enum ne_sock_sversion version;
+        int addr;
+        const char *fqdn;
+        unsigned int port;
+        const char *username, *password;
+    } ts[] = {
+        { NE_SOCK_SOCKSV4, 4, NULL, 55555, NULL, NULL },
+        { NE_SOCK_SOCKSV4, 4, NULL, 55555, "foobar", NULL },
+        { NE_SOCK_SOCKSV4A, 0, "www.example.com", 55555, NULL, NULL },
+        { NE_SOCK_SOCKSV5, 0, "www.example.com", 55555, NULL, NULL },
+        { NE_SOCK_SOCKSV5, 4, NULL, 55555, NULL, NULL },
+#ifdef TEST_IPV6
+        { NE_SOCK_SOCKSV5, 6, NULL, 55555, NULL, NULL },
+#endif
+        { NE_SOCK_SOCKSV5, 0, "www.example.com", 55555, "norman", "foobar" }
+    };
+    unsigned n;
+
+    for (n = 0; n < sizeof(ts)/sizeof(ts[n]); n++) {
+        ne_socket *sock;
+        struct socks_server arg = {0};
+        int ret;
+
+        arg.version = ts[n].version;
+        arg.expect_port = ts[n].port;
+        if (ts[n].addr == 4)
+            arg.expect_addr = ne_iaddr_make(ne_iaddr_ipv4, raw_127);
+#ifdef TEST_IPV6
+        else if (ts[n].addr == 6)
+            arg.expect_addr = ne_iaddr_make(ne_iaddr_ipv4, raw6_cafe);
+#endif
+        else
+            arg.expect_fqdn = ts[n].fqdn;
+        arg.username = ts[n].username;
+        arg.password = ts[n].password;
+        
+        CALL(begin_socks(&sock, &arg, echo_server, NULL));
+
+        ret = ne_sock_proxy(sock, ts[n].version, arg.expect_addr, 
+                            ts[n].fqdn, ts[n].port,
+                            ts[n].username, ts[n].password);
+        ONV(ret, ("proxy connect #%u gave %d", n, ret));
+        FULLREAD("ok!\n");
+        ECHO("hello,\n");
+        ECHO("\n");
+        ECHO("world\n");
+        
+        if (ts[n].addr)
+            ne_iaddr_free(arg.expect_addr);
+
+        CALL(finish(sock, 0));
+    }
+
+    return OK;
+}
+
+static int fail_socks(void)
+{
+    static const struct {
+        enum ne_sock_sversion version;
+        enum socks_failure failure;
+        const char *expect;
+        const char *username, *password;
+    } ts[] = {
+        { NE_SOCK_SOCKSV5, fail_init_vers, 
+          "Invalid version in proxy response", NULL, NULL },
+        { NE_SOCK_SOCKSV5, fail_init_trunc,
+          "Could not read initial response from proxy: Connection closed",
+          NULL, NULL },
+        { NE_SOCK_SOCKSV5, fail_init_close, 
+          "Could not read initial response from proxy: Connection closed", 
+          NULL, NULL },
+        { NE_SOCK_SOCKSV5, fail_no_auth, 
+          "No acceptable authentication method",
+          NULL, NULL },
+        { NE_SOCK_SOCKSV5, fail_bogus_auth, 
+          "Unexpected authentication method chosen",
+          NULL, NULL },
+        { NE_SOCK_SOCKSV5, fail_auth_close, 
+          "Could not read login reply: Connection closed",
+          "foo", "bar" },
+        { NE_SOCK_SOCKSV5, fail_auth_denied, 
+          "Authentication failed", "foo", "bar" }
+    };
+    unsigned n;
+
+    for (n = 0; n < sizeof(ts)/sizeof(ts[n]); n++) {
+        ne_socket *sock;
+        struct socks_server arg = {0};
+        int ret;
+
+        arg.version = ts[n].version;
+        arg.failure = ts[n].failure;
+        arg.expect_port = 5555;
+        arg.expect_addr = ne_iaddr_make(ne_iaddr_ipv4, raw_127);
+        arg.username = ts[n].username;
+        arg.password = ts[n].password;
+        
+        CALL(begin_socks(&sock, &arg, echo_server, NULL));
+
+        ret = ne_sock_proxy(sock, ts[n].version, arg.expect_addr, 
+                            NULL, arg.expect_port,
+                            ts[n].username, ts[n].password);
+        ONV(ret == 0, 
+            ("proxy connect #%u succeeded, expected failure '%s'", n, 
+             ts[n].expect));
+        
+        if (ret != 0 && strstr(ne_sock_error(sock), ts[n].expect) == NULL) {
+            t_warning("proxy connect #%u got unexpected failure '%s', wanted '%s'",
+                      n, ne_sock_error(sock), ts[n].expect);
+        }    
+
+        ne_iaddr_free(arg.expect_addr);
+
+        CALL(finish(sock, 0));
+    }
+
+    return OK;
+}
 
 ne_test tests[] = {
     T(multi_init),
@@ -1012,10 +1467,14 @@ ne_test tests[] = {
     T_LEAKY(init_ssl),
 #endif
     T(addr_make_v4),
+    T(parse_v4),
     T(addr_make_v6),
     T(addr_compare),
+    T(addr_reverse),
     T(just_connect),
     T(addr_connect),
+    T(addr_peer),
+    T(addr_canonical),
     T(read_close),
     T(peek_close),
     T(single_read),
@@ -1024,6 +1483,7 @@ ne_test tests[] = {
     T(read_and_peek),
     T(larger_read),
     T(ssl_session_id),
+    T(cipher),
     T(line_simple),
     T(line_closure),
     T(line_empty),
@@ -1034,19 +1494,27 @@ ne_test tests[] = {
     T(line_long_chunked),
     T(small_writes),
     T(large_writes),
+    T(large_writev),
     T(echo_lines),
     T(blocking),
+    T(prebind),
+    T(error),
 #ifdef SOCKET_SSL
     T(ssl_closure),
     T(ssl_truncate),
 #else
     T(write_reset),
     T(read_reset),
+#endif
+#if TEST_CONNECT_TIMEOUT
+    T(connect_timeout),
 #endif
     T(read_timeout),
     T(peek_timeout),
     T(readline_timeout),
     T(fullread_timeout),
     T(block_timeout),
+    T(socks_proxy),
+    T(fail_socks),
     T(NULL)
 };
